# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass

@dataclass
class LPType:
    
    HYBRID = 'HYBRID'
    SYNTHETIC = 'SYNTHETIC'
